# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


from cProfile import label
import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as dtparse
import numpy as np
import re
import requests
import bs4
import os
import pathlib

from geospacelab.config import prf
import geospacelab.datahub.sources.madrigal as madrigal
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool


def test():
    dt_fr = datetime.datetime(2016, 3, 14)
    dt_to = datetime.datetime(2016, 3, 14)
    download_obj = Downloader(dt_fr, dt_to, dry_run=False)


file_type_dict = {
    'combined': 'combined',
    'zenith single pulse': 'zenith single-pulse',
    'zenith alternating': 'zenith alternating',
    'misa single-pulse': 'misa (steerable) single-pulse',
    'misa alternating': 'misa (steerable) alternating',
    'gridded': 'gridded',
    'ion velocity': 'ion velocities',
}


class Downloader(object):
    """Download the Millstone ISR data
    """

    def __init__(self, dt_fr, dt_to, data_file_root_dir=None, file_type='combined', exp_name_pattern='',
                 dry_run=False,
                 user_fullname=madrigal.default_user_fullname,
                 user_email=madrigal.default_user_email,
                 user_affiliation=madrigal.default_user_affiliation):
        """

        :param dt_fr: Starting time.
        :type dt_fr: datetime.datetime
        :param dt_to: Stopping time.
        :type dt_to: datetime.datetime
        :param data_file_root_dir: The root directory for the data.
        :type data_file_root_dir: str  or pathlib.Path
        :param file_type: The file type associated with the data product.
        :type file_type: str, {'basic', 'zenith single-pulse', 'zenith alternating',
            'misa (steerable) single-pulse', 'misa (steerable) alternating', 'gridded', 'ion-velocities'
        :param experiment_key:
        :param user_fullname:
        :param user_email:
        :param user_affiliation:
        """
        self.user_fullname = user_fullname
        self.user_email = user_email
        self.user_affiliation = user_affiliation

        dt_fr = dttool.get_start_of_the_day(dt_fr)
        dt_to = dttool.get_end_of_the_day(dt_to)
        self.file_type = file_type
        self.exp_name_pattern = exp_name_pattern
        self.dry_run = dry_run
        if dt_fr == dt_to:
            dt_to = dt_to + datetime.timedelta(hours=23, minutes=59)
        self.dt_fr = dt_fr  # datetime from
        self.dt_to = dt_to  # datetime to

        if self.file_type not in file_type_dict.keys():
            raise TypeError("The file type cannot be identified!")

        if data_file_root_dir is None:
            self.data_file_root_dir = prf.datahub_data_root_dir / 'Madrigal' / 'Millstone_ISR'
        else:
            self.data_file_root_dir = pathlib.Path(data_file_root_dir)
        self.done = False

        self.madrigal_url = "http://millstonehill.haystack.mit.edu/"
        self.download_madrigal_files()

    def download_madrigal_files(self, download_pp=False):
        icodes = [30, ]
        all_files=[]
        for icode in icodes:
            mylog.simpleinfo.info("Searching data from the Madrigal database ...")
            exp_list, _, database = madrigal.utilities.list_experiments(
                icode, self.dt_fr, self.dt_to, madrigal_url=self.madrigal_url
            )
                    
            if self.dry_run:
                mylog.simpleinfo.info("Dry run mode: check the online file information...")
                mylog.simpleinfo.info("Listing all of the online files:")
                mylog.simpleinfo.info("{:<40s}\t{:<20s}{:<30s}{:<80s}".format(
                    'Experiment Name', 'Experiment Label', 'File Name', 'File Type'
                    )
                )
                for exp in exp_list:
                    files = database.getExperimentFiles(exp.id)
                    for file in files:
                        all_files.append([exp, file])
                for exp, file in all_files:
                    exp_url = exp.url
                    res = re.search(r'mlh/([\d]{2}[a-z]{3}[\d]{2}[\w]*)', exp_url)
                    exp_label = res.groups()[0]
                    file_path_remote = pathlib.Path(file.name)
                    file_name = file_path_remote.name 
                    mylog.simpleinfo.info(
                        "{:<40.40s}\t{:<20.20s}{:<30.30s}{:<80.80s}".format(
                            exp.name,
                            exp_label,
                            file_name,
                            file.kindatdesc, 
                        ) 
                    )
                return 
            
            matches = ['usrp receiver', 'calibration', 'f2 peak density']
            num_exp = 0
            
            for exp in exp_list:
                exp_url = exp.url
                res = re.search(r'mlh/([\d]{2}[a-z]{3}[\d]{2}[\w]*)', exp_url)
                exp_label = res.groups()[0]
                if not str(self.exp_name_pattern):
                    if any(x in exp.name.lower() for x in matches):
                        continue
                elif self.exp_name_pattern not in exp.name.lower():
                    continue
                
                mylog.simpleinfo.info(f"Searching files from the experiment {exp.name} ...")
                
                num_exp = num_exp + 1
                files = database.getExperimentFiles(exp.id)
                num_file = 0
                for file in files:
                    if file_type_dict[self.file_type] not in file.kindatdesc.lower():
                        continue
                    
                    num_file = num_file + 1

                    file_path_remote = pathlib.Path(file.name)
                    file_name = file_path_remote.name
                    res = re.search(r'/([\d]{2}[a-z]{3}[\d]{2})', file.doi)
                    dtstr = res.groups()[0]
                    dtstr = dtstr[0:2] + dtstr[2].upper() + dtstr[3:]
                    thisday = datetime.datetime.strptime(dtstr, "%d%b%y")

                    exp_name_patterns = re.findall('\w+', exp.name)
                    if len(exp_name_patterns) > 5:
                        exp_name_patterns = exp_name_patterns[:5]
                    exp_name = '-'.join(exp_name_patterns).lower()
                    data_file_dir = self.data_file_root_dir / thisday.strftime("%Y") / thisday.strftime('%Y%m%d') / exp_name
                    data_file_dir.mkdir(parents=True, exist_ok=True)
                    file_name_new = 'MillstoneHill_ISR_' + self.file_type.replace(' ', '_') + '_' + thisday.strftime('%Y%m%d') + \
                                    '.' + file_name.split('.')[-2] + '.' + file_name.split('.')[-1]
                    data_file_path = data_file_dir / file_name_new
                    if data_file_path.is_file():
                        mylog.simpleinfo.info("The file {} has been downloaded.".format(data_file_path.name))
                        self.done = True
                        continue

                    mylog.simpleinfo.info("Downloading  {} from the Madrigal database ...".format(file_name))
                    database.downloadFile(
                        file.name, data_file_path,
                        self.user_fullname, self.user_email, self.user_affiliation,
                        "hdf5"
                    )
                    self.done = True
                    mylog.simpleinfo.info("Save the file as {} in the directory {}".format(file_name_new, data_file_dir))
                if num_file > 1:
                    mylog.StreamLogger.warning("Multiple files matched and downloaded! Check the file information, please!")
            if num_exp > 1:
                mylog.StreamLogger.warning("Multiple experiments matched! Check the experiment information, please!") 

if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu


from geospacelab.datahub import DatabaseModel


class TUDDatabase(DatabaseModel):
    def __new__(cls, str_in, **kwargs):
        obj = super().__new__(cls, str_in, **kwargs)
        return obj


tud_database = TUDDatabase('TUDelft/Themosphere')
tud_database.url = 'http://thermosphere.tudelft.nl/'
tud_database.category = 'online database'
tud_database.Notes = """

References

If you use the data on this webpage, please cite our publications. 

Accelerometer-derived density data

G. March, J. van den IJssel, C. Siemes, P. Visser, E. Doornbos, M. Pilinski (2021) Gas-surface interactions modelling influence on satellite aerodynamics and thermosphere density. Journal of Space Weather and Space Climate, 11: 54. https://doi.org/10.1051/swsc/2021035

C. Siemes, J. de Teixeira da Encarnação, E. N. Doornbos, J. van den IJssel, J. Kraus, R. Pereštý, L. Grunwaldt, G. Apelbaum, J. Flury and P. E. Holmdahl Olsen (2016) Swarm accelerometer data processing from raw accelerations to thermospheric neutral densities. Earth, Planets and Space, 68: 92. https://doi.org/10.1186/s40623-016-0474-5

E. N. Doornbos (2012) Thermosphere density and wind determination from satellite dynamics. Dissertation, Springer Nature, Switzerland. ISBN 978-3-642-25129-0 (free PDF version)


Satellite geometry and aerodynamic models

G. March, E. N. Doornbos and P. N. M. A. Visser (2019) High-fidelity geometry models for improving the consistency of CHAMP, GRACE, GOCE and Swarm thermospheric density data sets. Advances in Space Research 63: 213–238. https://doi.org/10.1016/j.asr.2018.07.009

G. March, E. N. Doornbos and P. N. M. A. Visser (2019) CHAMP and GOCE thermospheric wind characterization with improved gas-surface interactions modelling. Advances in Space Research 64: 1225–1242. https://doi.org/10.1016/j.asr.2019.06.023


GNSS-derived density data

J. van den IJssel, E. Doornbos, E. Iorfida, G. March, C. Siemes, O. Montenbruck (2020) Thermosphere densities derived from Swarm GPS observations. Advances in Space Research. https://doi.org/10.1016/j.asr.2020.01.004

P. Visser and J. van den IJssel J (2016) Orbit determination and estimation of non-gravitational accelerations for the GOCE reentry phase. Advances in Space Research 58: 1840–1853. https://doi.org/10.1016/j.asr.2016.07.013


The original data was obtained from the following sources

CHAMP data: GFZ's Information System and Data Center
GOCE and Swarm data: ESA's Earth Online webpage
GRACE data: NASA's Physical Oceanography Distributed Active Archive Center (PO.DAAC)

"""


__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"

import datetime
import h5py

from geospacelab.datahub.sources.madrigal.isr.risr_n.downloader import Downloader as DownloaderBase
from geospacelab.config import prf



INCLUDE_FILE_TYPE_PATTERNS = [['velocity']]


class Downloader(DownloaderBase):

    def __init__(
            self, dt_fr: datetime.datetime, dt_to: datetime,
            include_exp_name_patterns: list=None,
            exclude_exp_name_patterns: list=None,
            include_exp_ids: list = None,
            exclude_exp_ids: list= None,
            include_file_name_patterns: list = None,
            exclude_file_name_patterns: list = None,
            include_file_type_patterns=None,
            exclude_file_type_patterns=None,
            data_file_root_dir: str = None,
            direct_download = True,
            force_download = False,
            dry_run: bool=False,
            madrigal_url: str = "https://data.amisr.com/madrigal"):
        exclude_exp_ids = [100213840, ] if exclude_exp_ids is None else exclude_exp_ids
        include_file_type_patterns = INCLUDE_FILE_TYPE_PATTERNS if include_file_type_patterns is None else include_file_type_patterns
        data_product = 'vi'
        data_file_root_dir = prf.datahub_data_root_dir / 'Madrigal' / 'RISR-N' if data_file_root_dir is None else data_file_root_dir
        super().__init__(
            dt_fr=dt_fr, dt_to=dt_to,
            data_product=data_product,
            include_exp_name_patterns=include_exp_name_patterns,
            exclude_exp_name_patterns=exclude_exp_name_patterns,
            include_exp_ids=include_exp_ids,
            exclude_exp_ids=exclude_exp_ids,
            include_file_name_patterns=include_file_name_patterns,
            exclude_file_name_patterns=exclude_file_name_patterns,
            include_file_type_patterns=include_file_type_patterns,
            exclude_file_type_patterns=exclude_file_type_patterns,
            data_file_root_dir=data_file_root_dir,
            force_download=force_download, direct_download=direct_download, dry_run=dry_run,
            madrigal_url=madrigal_url,)
        
    def download(self, **kwargs):
        
        file_paths = super().download(**kwargs)
        
        return file_paths
            
            





# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import logging.config

config = {
    'version': 1,
    'formatters': {
        'simple': {
            'format': '%(message)s',
        },
        'normal': {
            'format': '%(levelname)s: %(message)s',
        },
        'full': {
            'format': '%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        }
        # ohter formatter
    },
    'handlers': {
        'console-simple': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'simple'
        },
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'formatter': 'normal'
        },
        'file': {
            'class': 'logging.FileHandler',
            'filename': 'logging.log',
            'level': 'DEBUG',
            'formatter': 'full'
        },
        # other handler
    },
    'loggers':{
        'StreamLogger-simple': {
            'handlers': ['console-simple'],
            'level': 'DEBUG',
        },
        'StreamLogger': {
            'handlers': ['console'],
            'level': 'DEBUG',
        },
        'FileLogger': {
            #  console Handler and file Handler
            'handlers': ['console', 'file'],
            'level': 'DEBUG',
        },
        # other Logger
    }
}

logging.config.dictConfig(config)
simpleinfo = logging.getLogger("StreamLogger-simple")
StreamLogger = logging.getLogger("StreamLogger")
FileLogger = logging.getLogger("FileLogger")




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import matplotlib as mpl
import matplotlib.pyplot as plt

from geospacelab.visualization.mpl.dashboards import Dashboard


def test():
    myfig = plt.figure()
    move_figure((500, 50))
    set_figure_size([15, 15])
    plt.show()


class Figure(plt.Figure):

    def __init__(self, *args, **kwargs):
        """
        custom kwargs:
        figtitle: figure title added on the top-center
        fignote: figure note added on the right conner
        figsizeunit: 'centimeters' or 'inches', call set_figsize
        figposition: (x,y) starting from the left-upper conner of the screen, call move_figure
        """

        self.title = kwargs.pop('title', None)
        self.note = kwargs.pop('note', None)
        self.size = kwargs.pop('size', (10, 10))
        self.size_unit = kwargs.pop('size_unit', 'inches')
        self.position = kwargs.pop('position', (300, 100))
        self.dashboards = {}
        super().__init__(*args, **kwargs)

        self.set_figure_size(self.size, self.size_unit)

    def set_figure_size(self, size=None, unit="inches"):
        self.size = size
        self.size_unit = unit
        set_figure_size(size=size, unit=unit)

    def add_dashboard(self, dashboard=None, index=None, gs_num_rows=None, gs_num_cols=None, **kwargs):
        if dashboard is None:
            dashboard = Dashboard(**kwargs)
        elif not isinstance(dashboard, Dashboard):
            raise TypeError

        if index is None:
            index = len(self.dashboards.keys()) + 1

        if gs_num_rows is not None:
            dashboard.set_grid_layout(gs_num_rows, gs_num_cols)

        self.dashboards[index] = dashboard



        return index

    @property
    def position(self):
        return self._position

    @position.setter
    def position(self, value):
        if value is None:
            self._position = (300, 100)
        elif isinstance(value, tuple):
            self._position = value
            move_figure(self._position[0], self._position[1])
        else:
            raise TypeError("The position argument must be a 2-element tuple, e.g., (500, 200)!")


def set_figure_size(size=None, unit='centimeters', fig=None):
    if fig is None:
        fig = plt.gcf()
    if unit == 'centimeters':
        size[0] = size[0] / 2.54
        size[1] = size[1] / 2.54
    fig.set_size_inches(size[0], size[1], forward=True)
    return


def move_figure(x, y, fig=None):
    """Move figure's upper left corner to pixel (x, y)"""
    if fig is None:
        fig = plt.gcf()
    backend = mpl.get_backend()
    if backend == 'TkAgg':
        fig.canvas.manager.window.wm_geometry("+%d+%d" % (x, y))
    elif backend == 'WXAgg':
        fig.canvas.manager.window.SetPosition((x, y))
    else:
        # This works for QT and GTK
        # You can also use window.setGeometry
        try:
            fig.canvas.manager.window.move(x, y)
        except:
            print('Fail to set the figure position. Backend: ' + backend)


if __name__ == "__main__":
    test()



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

import numpy as np

import geospacelab.datahub as datahub
from geospacelab.datahub import DatabaseModel, FacilityModel, InstrumentModel, ProductModel
from geospacelab.datahub.sources.superdarn import superdarn_database
from geospacelab.config import prf
import geospacelab.toolbox.utilities.pybasic as basic
import geospacelab.toolbox.utilities.pylogging as mylog
from geospacelab.datahub.sources.superdarn.potmap.loader import Loader as default_Loader
import geospacelab.datahub.sources.superdarn.potmap.variable_config as var_config


default_dataset_attrs = {
    'database': superdarn_database,
    'facility': 'SuperDARN',
    'product': 'POTMAP',
    'data_file_ext': 'nc',
    'data_root_dir': prf.datahub_data_root_dir / 'SuperDARN' / 'PotentialMap',
    'allow_load': True,
    'allow_download': False,
    'data_search_recursive': False,
    'label_fields': ['database', 'facility', 'product'],
    'load_mode': 'assigned',
    'time_clip': True,
}

default_variable_names = [
    'DATETIME',
    'GRID_MLAT', 'GRID_MLON', 'GRID_MLT',
    'GRID_phi',
    'VCNUM', 'IMF_MODEL', 'CLOCK_ANGLE', 'DIP_TILT', 'E_SW',
    'SD_MODEL', 'FIT_ORDER', 'B_x_OMNI', 'B_y_OMNI', 'B_z_OMNI',
    'phi_CPCP', 'phi_MAX', 'phi_MIN'
    ]

# default_data_search_recursive = True

default_attrs_required = []


class Dataset(datahub.DatasetSourced):
    def __init__(self, **kwargs):
        kwargs = basic.dict_set_default(kwargs, **default_dataset_attrs)

        super().__init__(**kwargs)

        self.database = kwargs.pop('database', 'SuperDARN')
        self.facility = kwargs.pop('facility', 'SuperDARN')
        self.product = kwargs.pop('product', 'POTMAP')
        self.allow_download = kwargs.pop('allow_download', False)

        self.pole = kwargs.pop('pole', 'N')
        self.load_append_support_data = kwargs.pop('append_support_data', True)

        self.metadata = None

        allow_load = kwargs.pop('allow_load', False)

        # self.config(**kwargs)

        if self.loader is None:
            self.loader = default_Loader

        if self.downloader is None:
            self.downloader = None

        self._validate_attrs()

        if allow_load:
            self.load_data()

    def _validate_attrs(self):
        for attr_name in default_attrs_required:
            attr = getattr(self, attr_name)
            if not list(attr):
                mylog.StreamLogger.warning("The parameter {} is required before loading data!".format(attr_name))

    def label(self, **kwargs):
        label = super().label()
        return label

    def load_data(self, **kwargs):
        self.check_data_files(**kwargs)

        self._set_default_variables(
            default_variable_names,
            configured_variables=var_config.configured_variables
        )
        for file_path in self.data_file_paths:
            load_obj = self.loader(
                file_path, file_ext=self.data_file_ext, pole=self.pole,
                append_support_data=self.load_append_support_data
            )

            for var_name in self._variables.keys():
                value = load_obj.variables[var_name]
                self._variables[var_name].join(value)

            # self.select_beams(field_aligned=True)
        if self.time_clip:
            self.time_filter_by_range()

    def get_time_ind(self, ut):
        delta_sectime = [delta_t.total_seconds() for delta_t in (self['DATETIME'].value.flatten() - ut)]
        ind = np.where(np.abs(delta_sectime) == np.min(np.abs(delta_sectime)))[0][0]
        return ind

    def grid_phi(self, mlat_data, mlt_data, phi_data, mlt_res=0.2, mlat_res=0.5, interp_method='cubic'):
        import scipy.interpolate as si
        x = np.arange(0, 24, mlt_res)
        y = np.arange(50, 90, mlat_res)
        grid_x, grid_y = np.meshgrid(x, y, indexing='ij')

        xdata = np.vstack((mlt_data, mlt_data+24.))
        ydata = np.vstack((mlat_data, mlat_data))
        zdata = np.vstack((phi_data, phi_data))
        grid_phi = si.griddata(
            (xdata.flatten(), ydata.flatten()),
            zdata.flatten(),
            (grid_x, grid_y),
            method=interp_method,
            rescale=False
        )
        grid_mlat = grid_y
        grid_mlt = grid_x
        return grid_mlat, grid_mlt, grid_phi

    def postprocess_roll(self, mlat_data, mlt_data, phi_data):
        ind_min = np.argmin(mlt_data, axis=0)[0]
        mlat_data = np.roll(mlat_data, -ind_min, axis=0)
        mlt_data = np.roll(mlt_data, -ind_min, axis=0)
        phi_data = np.roll(phi_data, -ind_min, axis=0)
        return mlat_data, mlt_data, phi_data

    # def search_data_files(self, **kwargs):
    #     dt_fr = self.dt_fr
    #     if self.dt_to.hour > 22:
    #         dt_to = self.dt_to + datetime.timedelta(days=1)
    #     else:
    #         dt_to = self.dt_to
    #     diff_days = dttool.get_diff_days(dt_fr, dt_to)
    #     dt0 = dttool.get_start_of_the_day(dt_fr)
    #     for i in range(diff_days + 1):
    #         thisday = dt0 + datetime.timedelta(days=i)
    #         initial_file_dir = kwargs.pop('initial_file_dir', None)
    #         if initial_file_dir is None:
    #             initial_file_dir = self.data_root_dir / self.sat_id.lower() / thisday.strftime("%Y%m%d")
    #         file_patterns = [
    #             self.sat_id.upper(),
    #             self.product.upper(),
    #             thisday.strftime("%Y%m%d"),
    #         ]
    #         if self.orbit_id is not None:
    #             file_patterns.append(self.orbit_id)
    #         # remove empty str
    #         file_patterns = [pattern for pattern in file_patterns if str(pattern)]
    #
    #         search_pattern = '*' + '*'.join(file_patterns) + '*'
    #
    #         if self.orbit_id is not None:
    #             multiple_files = False
    #         else:
    #             fp_log = initial_file_dir / 'EDR-AUR.full.log'
    #             if not fp_log.is_file():
    #                 self.download_data(dt_fr=thisday, dt_to=thisday)
    #             multiple_files = True
    #         done = super().search_data_files(
    #             initial_file_dir=initial_file_dir,
    #             search_pattern=search_pattern,
    #             allow_multiple_files=multiple_files,
    #         )
    #         if done and self.orbit_id is not None:
    #             return True
    #
    #         # Validate file paths
    #
    #         if not done and self.allow_download:
    #             done = self.download_data(dt_fr=thisday, dt_to=thisday)
    #             if done:
    #                 done = super().search_data_files(
    #                     initial_file_dir=initial_file_dir,
    #                     search_pattern=search_pattern,
    #                     allow_multiple_files=multiple_files
    #                 )
    #
    #     return done
    #
    # def download_data(self, dt_fr=None, dt_to=None):
    #     if dt_fr is None:
    #         dt_fr = self.dt_fr
    #     if dt_to is None:
    #         dt_to = self.dt_to
    #     download_obj = self.downloader(
    #         dt_fr, dt_to,
    #         orbit_id=self.orbit_id, sat_id=self.sat_id, file_type=self.product.lower(),
    #         data_file_root_dir=self.data_root_dir)
    #     return download_obj.done

    @property
    def database(self):
        return self._database

    @database.setter
    def database(self, value):
        if isinstance(value, str):
            self._database = DatabaseModel(value)
        elif issubclass(value.__class__, DatabaseModel):
            self._database = value
        else:
            raise TypeError

    @property
    def product(self):
        return self._product

    @product.setter
    def product(self, value):
        if isinstance(value, str):
            self._product = ProductModel(value)
        elif issubclass(value.__class__, ProductModel):
            self._product = value
        else:
            raise TypeError

    @property
    def facility(self):
        return self._facility

    @facility.setter
    def facility(self, value):
        if isinstance(value, str):
            self._facility = FacilityModel(value)
        elif issubclass(value.__class__, FacilityModel):
            self._facility = value
        else:
            raise TypeError

    @property
    def instrument(self):
        return self._instrument

    @instrument.setter
    def instrument(self, value):
        if isinstance(value, str):
            self._instrument = InstrumentModel(value)
        elif issubclass(value.__class__, InstrumentModel):
            self._instrument = value
        else:
            raise TypeError



# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import h5py
import numpy as np
import cftime
import re
import datetime

import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.toolbox.utilities.pydatetime as dttool

pulse_code_dict = {
    'alternating': 97,
    'barker': 98,
    'single pulse': 115,
}

antenna_code_dict = {
    'zenith': 32,
    'misa': 31,
}

antenna_code_dict_r = {v: k for k, v in antenna_code_dict.items()}
pulse_code_dict_r = {v: k for k, v in pulse_code_dict.items()}

var_name_dict = {
    'AZ1': 'az1',
    'AZ2': 'az2',
    'EL1': 'el1',
    'EL2': 'el2',
    'PULSE_LENGTH': 'pl',
    'T_SYS': 'systmp',
    'POWER_NORM': 'pnrmd',
    'P_Tx': 'power',
    'MODE_TYPE': 'mdtyp',
    'POWER_LENGTH_F': 'pulf',
    'LAG_SPACING': 'dtau',
    'IPP': 'ipp',
    'f_Tx': 'tfreq',
    'v_PHASE_Tx': 'vtx',
    'v_PHASE_Tx_err': 'dvtx',
    'SCAN_TYPE': 'scntyp',
    'CYCN': 'cycn',
    'POSN': 'posn',
    'RANGE_RES': 'mresl',
    'RANGE': 'range',
    'SNR': 'snp3',
    'RESIDUAL': 'wchsq',
    'STATUS': 'gfit',
    'FIT_TYPE': 'fit_type',
    'FPI_QUALITY': 'fpi_dataqual',
    'ACF_NORM': 'fa',
    'ACF_NORM_ERR': 'dfa',
    'n_pp': 'popl',
    'n_pp_err': 'dpopl',
    'n_e': 'ne',
    'n_e_err': 'dne',
    'T_i': 'ti',
    'T_i_err': 'dti',
    'T_r': 'tr',
    'T_r_err': 'dtr',
    'nu_i': 'co',
    'nu_i_err': 'dco',
    'v_i_los': 'vo',
    'v_i_los_err': 'dvo',
    'comp_H_p': 'ph+',
    'comp_H_p_err': 'dph+',
    'comp_mix': 'pm',
    'comp_mix_err': 'dpm',
    'v_DOP_los': 'vdopp',
    'v_DOP_los_err': 'dvdopp',
    'HEIGHT': 'gdalt'
}


class Loader:
    """
    :param file_path: the file's full path
    :type file_path: pathlib.Path object
    :param file_type: the specific file type for the file being loaded. Options: ['TEC-MAT'], 'TEC-LOS', 'TEC-sites')
    :type file_type: str
    :param load_data: True, load without calling the method "load_data" separately.
    :type load_data: bool
    """
    def __init__(self, file_path, antenna='', pulse_code='', pulse_length=0, load_data=True):
        self.file_path = file_path
        self.antenna = antenna
        self.pulse_code = pulse_code
        self.pulse_length = pulse_length
        self.variables = {}
        self.metadata = {}

        self.done = False
        if load_data:
            self.load_from_table_layout()


    def load_from_table_layout(self):
        variables = {}

        with h5py.File(self.file_path, 'r') as fh5:
            table = fh5['Data']['Table Layout'][::]
            table_dtype = table.dtype

            # Get variable names
            table_var_names = list(table_dtype.fields.keys())
            # Get table data
            table_data = list(zip(*table))
            table_vars = {k: np.array(table_data[table_var_names.index(k)]) for k in table_var_names}
            # Check modes
            antenna_codes = table_vars['kinst'] = table_vars['kinst'].astype(np.int32)
            pulse_lengths = table_vars['pl'] = table_vars['pl'] * 1e6
            pulse_codes = table_vars['mdtyp'] = table_vars['mdtyp'].astype(np.int32)
            modes_ = list(zip(antenna_codes, pulse_codes, pulse_lengths))
            modes = np.empty(len(modes_), dtype=object)
            modes[:] = modes_
            modes_unique = self.get_modes(modes)

            mode = self.validate_mode(modes_unique)

            inds_mode = np.where((antenna_codes==mode[0]) & (pulse_codes==mode[1]) & (pulse_lengths==mode[2]) )[0]

            self.metadata['antenna'] = antenna_code_dict_r[mode[0]]
            self.metadata['pulse_code'] = pulse_code_dict_r[mode[1]]
            self.metadata['pulse_length'] = mode[2]

            # Grid data
            ut_unix_1 = table_vars['ut1_unix'][inds_mode]
            ut_unix_1_unique = np.unique(ut_unix_1)
            num_ut = len(ut_unix_1_unique)
            vars_fh5 = {}
            if len(ut_unix_1) % num_ut == 0:
                num_gate = int(len(ut_unix_1) / num_ut)

                for k, v in table_vars.items():
                    vars_fh5[k] = v[inds_mode].reshape((num_ut, num_gate))
            else:
                gate_inds = []
                gate_nums = []
                for t in ut_unix_1_unique:
                    ii = np.where(ut_unix_1==t)[0]
                    ran = table_vars['range'][inds_mode][ii]
                    diff_ran = np.diff(ran)
                    if any(diff_ran < 0): # duplicated ranges
                        iii = range(np.where(diff_ran<0)[0][0]+1)
                        ii = ii[iii]
                    gate_inds.append(ii)

                    gate_nums.append(len(gate_inds[-1]))
                max_gate_num = np.max(gate_nums)

                for k, v in table_vars.items():
                    vars_fh5[k] = np.empty((num_ut, max_gate_num))
                    vars_fh5[k][::] = np.nan
                    for i, inds in enumerate(gate_inds):
                        vars_fh5[k][i, 0:len(inds)] = v[inds_mode][inds]
            
            # Assign data
            records = fh5['Metadata']['_record_layout'][0]
            rec_var_names = np.array(list(records.dtype.fields.keys()))
            rec_vars = {str(rec_var_names[i]): int(records[i]) for i in range(len(rec_var_names))}
            for var_name, var_name_fh5 in var_name_dict.items():
                if var_name_fh5 not in vars_fh5.keys():
                    mylog.StreamLogger.warning(f"The requested variable {var_name_fh5} does not exist in the data file!")
                    variables[var_name] = None
                    continue
                if rec_vars[var_name_fh5] == 1:
                    variables[var_name] = vars_fh5[var_name_fh5][:, 0][:, np.newaxis]
                else:
                    variables[var_name] = vars_fh5[var_name_fh5]

            variables['comp_O_p'] = 1. - variables['comp_mix'] - variables['comp_H_p']
            variables['comp_O_p_err'] = np.sqrt(variables['comp_mix_err']**2 + variables['comp_H_p_err']**2)

            # need to be check when AZ close to 0.
            variables['AZ1'] = variables['AZ1'] % 360.
            variables['AZ2'] = variables['AZ2'] % 360.
            variables['AZ'] = (variables['AZ1'] + variables['AZ2']) / 2
            diff_az = np.abs(variables['AZ1'] - variables['AZ2'])
            variables['AZ'] = np.where(diff_az<180, variables['AZ'], ((variables['AZ1'] + variables['AZ2'] + 360) / 2) % 360)
            variables['EL'] = (variables['EL1'] + variables['EL2']) / 2 
            
            variables['DATETIME_1'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['ut1_unix'][:, 0])[:, np.newaxis]
            variables['DATETIME_2'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['ut2_unix'][:, 0])[:, np.newaxis] 
            variables['DATETIME'] = variables['DATETIME_1'] + (variables['DATETIME_2'] - variables['DATETIME_1']) / 2 
            variables['T_e'] = variables['T_i'] * variables['T_r']
            variables['T_e_err'] = variables['T_e'] * np.sqrt((variables['T_i_err'] / variables['T_i']) ** 2
                                                              + (variables['T_r_err'] / variables['T_r']) ** 2)
        self.variables = variables
         
        return

    def validate_mode(self, modes_unique):
        mode = None

        # Check antenna
        modes_antenna_matched = []
        try:
            for m in modes_unique:
                if m[0] == antenna_code_dict[self.antenna]:
                    modes_antenna_matched.append(m)
            if len(modes_antenna_matched) == 0:
                raise KeyError
            elif len(modes_antenna_matched) == 1:
                if str(self.pulse_code):
                    if pulse_code_dict[self.pulse_code] != modes_antenna_matched[0][1]:
                        raise ValueError
                if self.pulse_length > 0:
                    if self.pulse_length != modes_antenna_matched[0][2]:
                        raise ValueError
                return modes_antenna_matched[0]
            else:
                pass
        except Exception as e:
            if str(self.antenna):
                mylog.StreamLogger.error("Antenna {} is not found!".format(self.antenna.upper()))
            else:
                mylog.StreamLogger.error("Antenna must be specified!")
            self.list_modes()
            return None

        modes_pulse_code_matched = []
        try:
            for m in modes_antenna_matched:
                if m[1] == pulse_code_dict[self.pulse_code]:
                    modes_pulse_code_matched.append(m)
            if len(modes_pulse_code_matched) == 0:
                raise KeyError
            elif len(modes_pulse_code_matched) == 1:
                if self.pulse_length > 0:
                    if self.pulse_length != modes_antenna_matched[0][2]:
                        raise ValueError
                return modes_pulse_code_matched[0]
            else:
                pass
        except Exception as e:
            if str(self.pulse_code):
                mylog.StreamLogger.error("Pulse code {} is not found!".format(self.pulse_code.upper()))
            else:
                mylog.StreamLogger.error("Pulse code must be specified!")
            self.list_modes()
            return None

        modes_pulse_length_matched = []
        try:
            for m in modes_pulse_code_matched:
                if m[2] == self.pulse_length:
                    modes_pulse_length_matched.append(m)
            if len(modes_pulse_length_matched) == 0:
                raise ValueError
            elif len(modes_pulse_length_matched) == 1:
                return modes_pulse_length_matched[0]
            else:
                mylog.StreamLogger.error("Multiple modes found!")
                raise ValueError
        except Exception as e:
            if self.pulse_length > 0:
                mylog.StreamLogger.error("Pulse length {} is not found!".format(self.pulse_code.upper()))
            else:
                mylog.StreamLogger.error("Pulse length must be specified!")
            self.list_modes()
            return None
        return mode

    def list_modes(self):
        mylog.simpleinfo.info("List of the experiment modes:")

        for i, m in enumerate(self.metadata['MODES']):
            s = '{}, '.format(i)
            for k, v in m.items():
                s = s + "{}: {}, ".format(k, v)
            mylog.simpleinfo.info(s)
        return
    def get_modes(self, modes):

        modes_unique = np.unique(modes)

        self.metadata['MODES'] = []
        num_modes = self.metadata['NUM_MODES'] = len(modes_unique)
        for i in range(num_modes):
            self.metadata['MODES'].append({
                'ANTENNA_ID': modes_unique[i][0],
                'ANTENNA': antenna_code_dict_r[modes_unique[i][0]],
                'PULSE_CODE_ID': modes_unique[i][1],
                'PULSE_CODE': pulse_code_dict_r[modes_unique[i][1]],
                'PULSE_LENGTH': modes_unique[i][2]
            })
        return modes_unique

    def load(self):
        variables = {}
        metadata = {}
        with h5py.File(self.file_path, 'r') as fh5:
            data_fh5 = fh5['Data']
            array_layouts = list(data_fh5['Array Layout'].keys())

            antenna_codes = []
            pulse_codes = []
            pulse_lengths = []
            match = 0
            ind_array = None
            for i, array_layout_str in enumerate(array_layouts):
                res = re.search(r'kinst=([\d.]+).*mdtyp=([\d.]+).*pl=([\d.]+)', array_layout_str)
                antenna_code = int(float(res.groups()[0]))
                pulse_code = int(float(res.groups()[1]))
                pulse_length = float(res.groups()[2]) * 1e6
                antenna_codes.append(antenna_code)
                pulse_codes.append(pulse_code)
                pulse_lengths.append(pulse_length)
                if not str(self.antenna) or not str(self.pulse_code) or self.pulse_length == 0:
                    continue
                if antenna_code != antenna_code_dict[self.antenna] or \
                        pulse_code != pulse_code_dict[self.pulse_code] or \
                        pulse_length != self.pulse_length:
                    continue
                match = match + 1
                ind_array = i
            antenna_code_dict_r = {v: k for k, v in antenna_code_dict.items()}
            pulse_code_dict_r = {v: k for k, v in pulse_code_dict.items()}
            if match != 1:
                mylog.StreamLogger.error("The inputs do not match the Array Layouts! Check the Array Layout info below:")
                mylog.simpleinfo.info(
                        "{:10s}{:20s}{:30s}{:30s}".format(
                            'No.',
                            'antenna',
                            'pulse_code',
                            'pulse_length',
                        )
                    )
                for ind, (antenna_code, pulse_code, pulse_length) \
                        in enumerate(zip(antenna_codes, pulse_codes, pulse_lengths)):

                    mylog.simpleinfo.info(
                        "{:<10d}{:20s}{:30s}{:<30f}".format(
                            ind+1,
                            antenna_code_dict_r[antenna_code],
                            pulse_code_dict_r[pulse_code],
                            pulse_length,
                        )
                    )
                raise ValueError("Check the message above!")

            metadata['antenna'] = antenna_code_dict_r[antenna_codes[ind_array]]
            metadata['pulse_code'] = pulse_code_dict_r[pulse_codes[ind_array]]
            metadata['pulse_length'] = pulse_lengths[ind_array]

            vars_fh5 = {}
            array_layout_str = array_layouts[ind_array]
            fh5_vars_1d = data_fh5['Array Layout'][array_layout_str]['1D Parameters']
            for var_name in fh5_vars_1d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_1d[var_name])[:, np.newaxis]
            fh5_vars_2d = data_fh5['Array Layout'][array_layout_str]['2D Parameters']
            for var_name in fh5_vars_2d.keys():
                if var_name == 'Data Parameters':
                    continue
                vars_fh5[var_name] = np.array(fh5_vars_2d[var_name]).T
            vars_fh5['range'] = np.array(data_fh5['Array Layout'][array_layout_str]['range'])[np.newaxis, :]
            vars_fh5['timestamps'] = np.array(data_fh5['Array Layout'][array_layout_str]['timestamps'])[:, np.newaxis]
            for var_name, var_name_fh5 in var_name_dict.items():
                if var_name_fh5 not in vars_fh5.keys():
                    mylog.StreamLogger.warning(f"The requested variable {var_name_fh5} does not exist in the data file!")
                    variables[var_name] = None
                    continue
                variables[var_name] = vars_fh5[var_name_fh5]

            variables['comp_O_p'] = 1. - variables['comp_mix'] - variables['comp_H_p']
            variables['comp_O_p_err'] = np.sqrt(variables['comp_mix_err']**2 + variables['comp_H_p_err']**2)

            # need to be check when AZ close to 0.
            variables['AZ1'] = variables['AZ1'] % 360.
            variables['AZ2'] = variables['AZ2'] % 360.
            variables['AZ'] = (variables['AZ1'] + variables['AZ2']) / 2
            diff_az = np.abs(variables['AZ1'] - variables['AZ2'])
            variables['AZ'] = np.where(diff_az<180, variables['AZ'], ((variables['AZ1'] + variables['AZ2'] + 360) / 2) % 360)
            variables['EL'] = (variables['EL1'] + variables['EL2']) / 2 

            variables['RANGE'] = np.tile(vars_fh5['range'], [variables['n_e'].shape[0], 1])
            variables['DATETIME'] = dttool.convert_unix_time_to_datetime_cftime(vars_fh5['timestamps'])
            variables['T_e'] = variables['T_i'] * variables['T_r']
            variables['T_e_err'] = variables['T_e'] * np.sqrt((variables['T_i_err'] / variables['T_i']) ** 2
                                                              + (variables['T_r_err'] / variables['T_r']) ** 2)

        self.variables = variables
        self.metadata = metadata




# Licensed under the BSD 3-Clause License
# Copyright (C) 2021 GeospaceLab (geospacelab)
# Author: Lei Cai, Space Physics and Astronomy, University of Oulu

__author__ = "Lei Cai"
__copyright__ = "Copyright 2021, GeospaceLab"
__license__ = "BSD-3-Clause License"
__email__ = "lei.cai@oulu.fi"
__docformat__ = "reStructureText"


import datetime
import numpy as np
import cftime
import requests
import bs4
import pathlib
import re
import netCDF4 as nc
import ftplib
from contextlib import closing

import geospacelab.toolbox.utilities.pydatetime as dttool
import geospacelab.toolbox.utilities.pylogging as mylog
import geospacelab.datahub.sources.gfz.downloader as downloader
import geospacelab.datahub.sources.wdc as wdc


class Downloader(downloader.Downloader):

    def __init__(self, data_file_root_dir=None, force=False, download_method='html'):

        ftp_sub_dir = 'Kp_ap_Ap_SN_F107'
        ftp_filename_prefix = ftp_sub_dir + '_'

        dt_fr = datetime.datetime.now()
        dt_to = datetime.datetime.now()
        super().__init__(
            dt_fr,  dt_to,
            data_file_root_dir=data_file_root_dir, force=force,
            ftp_sub_dir=ftp_sub_dir, ftp_filename_prefix=ftp_filename_prefix
        )
        self.data_file_root_dir = self.data_file_root_dir / 'Kp_Ap'
        self.download_method = download_method
        self.download()

    def download(self):
        if self.download_method == 'html':
            self.download_from_web()
        elif self.download_method == 'ftp':
            super().download()

    def download_from_web(self):
        mylog.StreamLogger.info("Downloading the nowcast data from the GFZ server ...")
        r = requests.get('https://www-app3.gfz-potsdam.de/kp_index/Kp_ap_Ap_SN_F107_nowcast.txt')
        r.raise_for_status()
        file_path = self.data_file_root_dir / 'Kp_ap_Ap_SN_F107_nowcast.txt'
        with open(file_path, "w") as file:
            file.write(r.text)
        self.save_to_netcdf(file_path)

    def save_to_netcdf(self, file_path):
        with open(file_path, 'r') as f:
            text = f.read()

            results = re.findall(
                r'^(\d+ \d+ \d+)\s*\d+\s*[\d.]+\s*(\d+)\s*(\d+)\s*' +
                r'([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*' +
                r'([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*([\-\d]+)\s*' +
                r'([\-\d]+)\s*([\-\d.]+)\s*([\-\d.]+)\s*([\-\d]+)',
                text,
                re.M
            )
            results = list(zip(*results))

            dts = [datetime.datetime.strptime(dtstr, "%Y %m %d") for dtstr in results[0]]
            time_array = np.array(cftime.date2num(dts, units='seconds since 1970-01-01 00:00:00.0'))
            bsr_array = np.array(results[1])
            bsr_array.astype(np.int32)
            db_array = np.array(results[2]).astype(np.int32)
            kp_array = np.array(results[3:11]).astype(np.float32)
            kp_array = np.where(kp_array == -1, np.nan, kp_array)
            ap_array = np.array(results[11:20]).astype(np.float32)
            ap_array = np.where(ap_array == -1, np.nan, ap_array)
            sn_array = np.array(results[20]).astype(np.int32)
            sn_array = np.where(sn_array == -1, np.nan, sn_array)
            f107o_array = np.array(results[21]).astype(np.float32)
            f107o_array = np.where(f107o_array == -1, np.nan, f107o_array)
            f107a_array = np.array(results[22]).astype(np.float32)
            f107a_array = np.where(f107a_array == -1, np.nan, f107a_array)
            flag_array = np.array(results[23]).astype(np.int32)

            num_rows = len(results[0])

            ################## for SN, f10.7
            fp = file_path.parent.resolve().parent.resolve() / "SN_F107" / ("GFZ_SN_F107_nowcast" + '.nc')
            fp.parent.resolve().mkdir(parents=True, exist_ok=True)
            fnc = nc.Dataset(fp, 'w')
            fnc.createDimension('UNIX_TIME', num_rows)

            fnc.title = "GFZ SN/F10.7 index"
            time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
            time.units = 'seconds since 1970-01-01 00:00:00.0'
            f107o = fnc.createVariable('F107_OBS', np.float32, ('UNIX_TIME',))
            f107a = fnc.createVariable('F107_ADJ', np.float32, ('UNIX_TIME',))
            sn = fnc.createVariable('SN', np.float32, ('UNIX_TIME',))
            bsr = fnc.createVariable('BSRN', np.float32, ('UNIX_TIME',))
            db = fnc.createVariable('BSRN_Days', np.float32, ('UNIX_TIME',))
            flag = fnc.createVariable('Flag', np.float32, ('UNIX_TIME',))
            time[::] = time_array[::]
            f107o[::] = f107o_array[::]
            f107a[::] = f107a_array[::]
            sn[::] = sn_array[::]
            bsr[::] = bsr_array[::]
            db[::] = db_array[::]
            flag_1 = np.where(flag_array < 2, 0, 1)
            flag[::] = flag_1[::]
            print('From {} to {}.'.format(
            datetime.datetime.utcfromtimestamp(time_array[0]),
            datetime.datetime.utcfromtimestamp(time_array[-1]))
            )
            mylog.StreamLogger.info(
                "The requested SN/F10.7 data has been downloaded and saved in the file {}.".format(fp))
            fnc.close()

            ########## for Kp Ap
            fp = file_path.parent.resolve() / ("GFZ_Kp_Ap_nowcast" + '.nc')
            fp.parent.resolve().mkdir(parents=True, exist_ok=True)
            fnc = nc.Dataset(fp, 'w')
            fnc.createDimension('UNIX_TIME', num_rows*8)

            fnc.title = "GFZ Kp/Ap index"
            time = fnc.createVariable('UNIX_TIME', np.float64, ('UNIX_TIME',))
            time.units = 'seconds since 1970-01-01 00:00:00.0'
            kp = fnc.createVariable('Kp', np.float32, ('UNIX_TIME',))
            ap = fnc.createVariable('ap', np.float32, ('UNIX_TIME',))
            Ap = fnc.createVariable('Ap', np.float32, ('UNIX_TIME',))
            flag = fnc.createVariable('flag', np.float32, ('UNIX_TIME',))
            seconds = np.arange(8) * 3600 * 3 + 1800*3
            time_array = np.tile(time_array, (8, 1)).T
            for i in range(8):
                time_array[:, i] = time_array[:, i] + seconds[i]
            time_array = time_array.flatten()
            kp_array = kp_array.T.flatten()
            Ap_array = np.tile(ap_array[-1, :].flatten(), (8, 1)).T.flatten()
            ap_array = ap_array[:-1, :].T.flatten()
            flag_array = np.tile(flag_array, (8, 1)).T.flatten()
            flag_2 = np.where(flag_array == 2, 1, 0)
            time[::] = time_array[::]
            kp[::] = kp_array[::]
            ap[::] = ap_array[::]
            Ap[::] = Ap_array[::]
            flag[::] = flag_2[::]
            print('From {} to {}.'.format(
            datetime.datetime.utcfromtimestamp(time_array[0]),
            datetime.datetime.utcfromtimestamp(time_array[-1]))
            )
            mylog.StreamLogger.info(
                "The requested Kp/Ap data has been downloaded and saved in the file {}.".format(fp))
            fnc.close()

            self.done = True


if __name__ == "__main__":
    dt_fr1 = datetime.datetime(1990, 1, 1)
    dt_to1 = datetime.datetime(2020, 12, 16)
    Downloader()



